import numpy as np


def pca(X, *Xtest, mincomponents=1, retvar=0.95):
    """Principal Component Analysis.

    Perform PCA dimensionality reduction.  The number of output
    components is the maximum between mincomponents and those required
    to ensure that at least a fraction retvar of the original variance
    is retained.

    Features are linearly transformed to have zero mean and null
    covariances.  Test features, when given, are processed using the
    transform estimated on X.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    Xtest : ndarray, shape (mtest, n)
         zero or more arrays of test features (one row per feature vector).
    mincomponents : int
         minimum number of output components.
    retvar : float
         minimum fraction of total variance retained in the output
         components.

    Returns
    -------
    ndarray, shape (m, output_n)
        normalized features.
    ndarray, shape (mtest, output_n)
        normalized test features (one for each array in Xtest).

    """
    # Compute the moments
    mu = X.mean(0, keepdims=True)
    sigma = np.cov(X.T)
    # Compute and sort the eigenvalues
    evals, evecs = np.linalg.eig(sigma)
    order = np.argsort(-evals)
    evals = evals[order]
    # Determine the components to retain
    k = 1 + (np.cumsum(evals) >= retvar * evals.sum()).nonzero()[0][0]
    k = max(k, mincomponents)
    w = evecs[:, order[:k]]  # 1e-15 avoids div. by zero
    # Transform the data
    X = (X - mu) @ w
    if not Xtest:
        return X
    Xtest = tuple((Xt - mu) @ w for Xt in Xtest)
    return (X,) + Xtest



import numpy as np


def multinomial_logreg_inference(X, W, b):
    """Predict class probabilities.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    W : ndarray, shape (n, k)
         weight vectors, each row representing a different class.
    b : ndarray, shape (k,)
         vector of biases.

    Returns
    -------
    P : ndarray, shape (m, k)
         probability estimates.
    """
    logits = X @ W + b.T
    return softmax(logits)


def softmax(Z):
    """Softmax operator.

    Parameters
    ----------
    Z : ndarray, shape (m, n)
         input vectors.

    Returns
    -------
    ndarray, shape (m, n)
         data after the softmax has been applied to each row.
    """
    # Subtracting the maximum improves numerical stability
    E = np.exp(Z - Z.max(1, keepdims=True))
    return E / E.sum(1, keepdims=True)


def one_hot_vectors(Y, classes):
    """Convert an array of labels into a matrix of one-hot vectors.

    Parameters
    ----------
    Y : ndarray, shape (m,)
         labels.
    classes : int
         number of classes.  If None it is deduced from Y.

    Returns
    -------
    ndarray, shape (m, classes)
         One-hot vectors representing the labels Y.
    """
    m = Y.shape[0]
    H = np.zeros((m, classes))
    H[np.arange(m), Y] = 1
    return H


def multinomial_logreg_train(X, Y, lambda_, lr=1e-3, steps=1000,
                             init_w=None, init_b=None):
    """Train a classifier based on multinomial logistic regression.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features.
    Y : ndarray, shape (m,)
        training labels with integer values in the range 0...(k-1).
    lambda_ : float
        regularization coefficient.
    lr : float
        learning rate
    steps : int
        number of training steps
    init_w : ndarray, shape (n, k)
        initial weights (None for zero initialization)
    init_b : ndarray, shape (k,)
        initial biases (None for zero initialization)

    Returns
    -------
    w : ndarray, shape (n, k)
        learned weights (one vector per class).
    b : ndarray, shape (k,)
        vector of biases.
    """
    m, n = X.shape
    k = Y.max() + 1
    W = (init_w if init_w is not None else np.zeros((n, k)))
    b = (init_b if init_b is not None else np.zeros(k))
    H = one_hot_vectors(Y, k)
    for step in range(steps):
        P = multinomial_logreg_inference(X, W, b)
        grad_W = (X.T @ (P - H)) / m + 2 * lambda_ * W
        grad_b = (P - H).mean(0)
        W -= lr * grad_W
        b -= lr * grad_b
    return W, b


def cross_entropy(H, P):
    """Average cross entropy.

    Parameters
    ----------
    H : ndarray, shape (m, k)
        one hot vectors for the target labels.
    P : ndarray, shape (m, k)
        probability estimates.

    Returns
    -------
    float
        average cross entropy.
    """
    return -(H * np.log(P)).sum(1).mean()



import numpy as np


def svm_inference(X, w, b):
    """SVM prediction of the class labels.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    w : ndarray, shape (n,)
         weight vector.
    b : float
         scalar bias.

    Returns
    -------
    ndarray, shape (m,)
        predicted labels (one per feature vector).
    ndarray, shape (m,)
        classification scores (one per feature vector).
    """
    logits = X @ w + b
    labels = (logits > 0).astype(int)
    return labels, logits


def hinge_loss(labels, logits):
    """Average hinge loss.

    Parameters
    ----------
    labels : ndarray, shape (m,)
        binary target labels (0 or 1).
    logits : ndarray, shape (m,)
        classification scores (logits).

    Returns
    -------
    float
        average hinge loss.
    """
    loss = np.maximum(0, 1 - (2 * labels - 1) * logits)
    return loss.mean()


def svm_train(X, Y, lambda_, lr=1e-3, steps=1000, init_w=None, init_b=0):
    """Train a binary SVM classifier.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features.
    Y : ndarray, shape (m,)
        binary training labels.
    lambda_ : float
        regularization coefficient.
    lr : float
        learning rate
    steps : int
        number of training steps
    init_w : ndarray, shape (n,)
        initial weights (None for zero initialization)
    init_b : float
        initial bias

    Returns
    -------
    w : ndarray, shape (n,)
        learned weight vector.
    b : float
        learned bias.
    """
    m, n = X.shape
    w = (init_w if init_w is not None else np.zeros(n))
    b = init_b
    C = (2 * Y) - 1
    for step in range(steps):
        labels, logits = svm_inference(X, w, b)
        hinge_diff = -C * ((C * logits) < 1)
        grad_w = (hinge_diff @ X) / m + lambda_ * w
        grad_b = hinge_diff.mean()
        w -= lr * grad_w
        b -= lr * grad_b
    return w, b



import numpy as np
import pvml


def _check_gradient(cnn, eps=1e-6, imsz=25, bsz=2):
    """Numerical check gradient computations."""
    insz = cnn.weights[0].shape[2]
    outsz = cnn.weights[-1].shape[3]
    X = np.random.randn(bsz, imsz, imsz, insz)
    Y = np.random.randint(0, outsz, (bsz,))
    A = cnn.forward(X)
    D = cnn.backward(Y, A)
    L = cnn.loss(Y, A[-1])
    for XX, W, DD, s in zip(A, cnn.weights, D, cnn.strides):
        grad_W = pvml.cnn._convolution_derivative(XX, DD, W.shape[0],
                                                  W.shape[1], s, s)
        GW = np.empty_like(W)
        for idx in np.ndindex(*W.shape):
            bak = W[idx]
            W[idx] += eps
            A1 = cnn.forward(X)
            L1 = cnn.loss(Y, A1[-1])
            W[idx] = bak
            GW[idx] = (L1 - L) / eps
        err = np.abs(GW - grad_W).max()
        print(err, "OK" if err < 1e-4 else "")
        assert err < 1e-4
    for b, DD in zip(cnn.biases, D):
        grad_b = DD.sum(2).sum(1).sum(0)
        Gb = np.empty_like(b)
        for idx in np.ndindex(*b.shape):
            bak = b[idx]
            b[idx] += eps
            A1 = cnn.forward(X)
            L1 = cnn.loss(Y, A1[-1])
            b[idx] = bak
            Gb[idx] = (L1 - L) / eps
        err = np.abs(Gb - grad_b).max()
        print(err, "OK" if err < 1e-4 else "")
        assert err < 1e-4


if __name__ == "__main__":
    cnn = pvml.CNN([4, 6], [2], [1])
    _check_gradient(cnn)
    # cnn = pvml.CNN([8, 7, 6], [1, 1], [1, 1])
    # _check_gradient(cnn)
    cnn = pvml.CNN([3, 8, 7, 5], [5, 4, 3], [2, 1, 1])
    _check_gradient(cnn)



import numpy as np


def ksvm_inference(X, Xtrain, alpha, b, kfun, kparam):
    """SVM prediction of the class labels.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    Xtrain : ndarray, shape (t, n)
         features used during training (one row per feature vector).
    alpha : ndarray, shape (t,)
         vector of learned coefficients.
    b : float
         scalar bias.
    kfun : string
         name of the kernel function
    kparam : float
         parameter of the kernel

    Returns
    -------
    ndarray, shape (m,)
        predicted labels (one per feature vector).
    ndarray, shape (m,)
        classification scores (one per feature vector).
    """
    K = kernel(X, Xtrain, kfun, kparam)
    logits = K @ alpha + b
    labels = (logits > 0).astype(int)
    return labels, logits


def kernel(X1, X2, kfun, kparam):
    """Compute the kernel between two groups of feature vectors.

    Parameters
    ----------
    X1 : ndarray, shape (m, n)
         first group of features (one row per feature vector).
    X2 : ndarray, shape (t, n)
         second group of features (one row per feature vector).
    kfun : string
         name of the kernel function
    kparam : float
         parameter of the kernel

    Returns
    -------
    K : ndarray, shape (m, t)
         matrix with the result of the kernel function applied to
         the two groups of feature vectors.
    """
    if kfun == "polynomial":
        return (X1 @ X2.T + 1) ** kparam
    elif kfun == "rbf":
        qx1 = (X1 ** 2).sum(1, keepdims=True)
        qx2 = (X2 ** 2).sum(1, keepdims=True)
        cross = 2 * X1 @ X2.T
        return np.exp(-kparam * (qx1 - cross + qx2.T))
    raise ValueError("Unknown kernel ('%s')" % kfun)


def ksvm_train(X, Y, kfun, kparam, lambda_, lr=1e-3, steps=1000,
               init_alpha=None, init_b=0):
    """Train a binary non-linear SVM classifier.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features.
    Y : ndarray, shape (m,)
        binary training labels.
    kfun : string
         name of the kernel function
    kparam : float
         parameter of the kernel
    lambda_ : float
        regularization coefficient.
    lr : float
        learning rate
    steps : int
        number of training steps
    init_alpha : ndarray, shape (m,)
        initial coefficients (None for zero initialization)
    init_b : float
        initial bias

    Returns
    -------
    alpha : ndarray, shape (m,)
        vector of learned coefficients.
    b : float
        learned bias.
    """
    K = kernel(X, X, kfun, kparam)
    m, n = X.shape
    alpha = (init_alpha if init_alpha is not None else np.zeros(m))
    b = (init_b if init_b is not None else 0)
    C = (2 * Y) - 1
    for step in range(steps):
        ka = K @ alpha
        logits = ka + b
        hinge_diff = -C * ((C * logits) < 1)
        grad_alpha = (hinge_diff @ K) / m + lambda_ * ka
        grad_b = hinge_diff.mean()
        alpha -= lr * grad_alpha
        b -= lr * grad_b
    return alpha, b



from .cnn import CNN as _CNN


_LAYERS = [
    (96, 7, 4),
    # (16, 1, 1),
    (128, 3, 1),
    (16, 1, 1),
    (128, 3, 1),
    (32, 1, 1),
    (256, 3, 2),
    (32, 1, 1),
    (256, 3, 1),
    (48, 1, 1),
    (384, 3, 1),
    (48, 1, 1),
    (384, 3, 1),
    (64, 1, 1),
    (512, 3, 2),
    (64, 1, 1),
    (512, 3, 1),
    (1000, 1, 1)
]


def make_pvmlnet(pretrained=False):
    channels = [3] + [x[0] for x in _LAYERS]
    kernels = [x[1] for x in _LAYERS]
    strides = [x[2] for x in _LAYERS]
    net = _CNN(channels, kernels, strides)
    return net


if __name__ == "__main__":
    import numpy as np
    net = make_pvmlnet(pretrained=False)
    parameters = sum(w.size for w in net.weights)
    parameters += sum(b.size for b in net.biases)
    print("{} parameters".format(parameters))
    images = np.empty((1, 224, 224, 3))
    activations = net.forward(images)
    for x in activations:
        print("x".join(map(str, x.shape)))



import numpy as np


def categorical_naive_bayes_train(X, Y, priors=None):
    """Train naive Bayes classifier for categorical data.

    Non integer feature values are truncated.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features with values in {0, ..., q - 1}.
    Y : ndarray, shape (m,)
        training labels with values in {0, ..., k - 1}.
    priors : ndarray, shape (k,)
        Prior probabilities for the classes (if None they get
        estimated from Y).

    Returns
    -------
    probs : ndarray, shape (k, n, q)
         class/feature probabilities.
    priors : ndarray, shape (k,)
         class prior probabilities.
    """
    X = np.maximum(X, 0).astype(int)  # Force X to be of non-negative integers
    m, n = X.shape
    q = X.max() + 1
    k = Y.max() + 1
    probs = np.empty((k, n, q))
    for c in range(k):
        indices = (Y == c).nonzero()[0]
        tot = indices.shape[0] + q  # With Laplacian smoothing
        for j in range(n):
            counts = (1 + np.bincount(X[indices, j], minlength=q))
            probs[c, j, :] = counts / tot
    if priors is None:
        priors = np.bincount(Y) / m
    return (probs, priors)


def categorical_naive_bayes_inference(X, probs, priors):
    """Categorical naive Bayes classifier inference.

    Non integer feature values are truncated.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features with values in {0, ..., q - 1}.
    probs : ndarray, shape (k, n, q)
        class mean vectors.
    priors : ndarray, shape (k,)
        class prior probabilities.

    Returns
    -------
    ndarray, shape (m,)
        predicted labels (one per feature vector).
    """
    q = probs.shape[2]
    X = np.clip(X.astype(int), 0, q - 1)
    m, n = X.shape
    k = priors.shape[0]
    scores = np.log(priors)[None, :].repeat(m, axis=0)
    for c in range(k):
        for j in range(n):
            scores[:, c] += np.log(probs[c, j, X[:, j]])
    labels = np.argmax(scores, 1)
    return labels, scores


def multinomial_naive_bayes_train(X, Y, priors=None):
    """Train a multinomial naive Bayes classifier.

    Non integer feature values are truncated.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features, they must be non-negative integers.
    Y : ndarray, shape (m,)
        training labels with values in {0, ..., k - 1}.
    priors : ndarray, shape (k,)
        Prior probabilities for the classes (if None they get
        estimated from Y).

    Returns
    -------
    W : ndarray, shape (n, k)
         weight matrix.
    b : ndarray, shape (k,)
         vector of biases.
    """
    X = np.maximum(X, 0).astype(int)  # Force X to be of non-negative integers
    m, n = X.shape
    k = Y.max() + 1
    probs = np.empty((k, n))
    for c in range(k):
        counts = X[Y == c, :].sum(0)
        tot = counts.sum()
        probs[c, :] = (counts + 1) / (tot + n)  # with Laplacian smoothing)
    if priors is None:
        priors = np.bincount(Y) / m
    W = np.log(probs).T
    b = np.log(priors)
    return (W, b)


def multinomial_naive_bayes_inference(X, W, b):
    """Multinomial naive Bayes classifier inference.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    W : ndarray, shape (n, k)
         weight vectors, each row representing a different class.
    b : ndarray, shape (k,)
         vector of biases.

    Returns
    -------
    ndarray, shape (m,)
        predicted labels (one per feature vector).
    """
    scores = X @ W + b.T
    labels = np.argmax(scores, 1)
    return labels, scores


def gaussian_naive_bayes_train(X, Y, priors=None):
    """Train a Gaussian Naive Bayes classifier.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features.
    Y : ndarray, shape (m,)
        training labels with values in {0, ..., k - 1}.
    priors : ndarray, shape (k,)
        Prior probabilities for the classes (if None they get
        estimated from Y).

    Returns
    -------
    means : ndarray, shape (k, n)
         class mean vectors.
    vars : ndarray, shape (k, n)
         per class variance of features.
    priors : ndarray, shape (k,)
         class prior probabilities.
    """
    k = Y.max() + 1
    m, n = X.shape
    means = np.empty((k, n))
    vars = np.empty((k, n))
    if priors is None:
        priors = np.bincount(Y) / m
    for c in range(k):
        indices = (Y == c).nonzero()[0]
        means[c, :] = X[indices, :].mean(0)
        vars[c, :] = X[indices, :].var(0)
    return (means, vars, priors)


def gaussian_naive_bayes_inference(X, means, vars, priors):
    """Gaussian Naive Bayes inference.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    means : ndarray, shape (k, n)
         class mean vectors.
    vars : ndarray, shape (k, n)
         variances.
    priors : ndarray, shape (k,)
         class prior probabilities.

    Returns
    -------
    ndarray, shape (m,)
        predicted labels (one per feature vector).
    """
    diffs = (X[:, None, :] - means[None, :, :]) ** 2
    diffs /= vars[None, :, :]
    scores = -0.5 * diffs.sum(2) - 0.5 * np.log(vars).sum(1)[None, :]
    scores += np.log(priors)[None, :]
    labels = np.argmax(scores, 1)
    return labels, scores



import numpy as np


def perceptron_train(X, Y, steps=10000, init_w=None, init_b=0):
    """Train a binary classifier using the perceptron algorithm.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features.
    Y : ndarray, shape (m,)
        binary training labels.
    steps: int
        maximum number of training iterations
    init_w : ndarray, shape (n,)
        initial weights (None for zero initialization)
    init_b : float
        initial bias

    Returns
    -------
    w : ndarray, shape (n,)
        learned weight vector.
    b : float
        learned bias.
    """
    w = (init_w if init_w is not None else np.zeros(X.shape[1]))
    b = init_b
    for step in range(steps):
        errors = 0
        for i in range(X.shape[0]):
            d = (1 if X[i, :] @ w + b > 0 else 0)
            w += (Y[i] - d) * X[i, :].T
            b += (Y[i] - d)
            errors += np.abs(Y[i] - d)
        if errors == 0:
            break
    return w, b


def perceptron_inference(X, w, b):
    """Perceptron prediction of the class labels.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    w : ndarray, shape (n,)
         weight vector.
    b : float
         scalar bias.

    Returns
    -------
    ndarray, shape (m,)
        predicted labels (one per feature vector).
    ndarray, shape (m,)
        classification scores (one per feature vector).
    """
    logits = X @ w + b
    labels = (logits > 0).astype(int)
    return labels, logits



import numpy as np


#!begin1
def one_vs_rest_train(X, Y, train_fun):
    """Train a multi-class classifier using the one vs. rest strategy.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features.
    Y : ndarray, shape (m,)
        training labels in the range 0, ..., (k - 1).
    train_fun : function that, given X and a set of binary labels, trains
        a binary classifier.  The function must return a tuple, where the
        last element is the list of loss values obtained during training,
        and all the others are the parameters of the trained classifier.

    Returns
    -------
    params : dict
        collection of (class, parameters) representing the binary classifiers.
    loss : ndarray, shape (steps,)
        average loss value after each training iteration.
    """
    k = Y.max() + 1
    m, n = X.shape
    loss = 0
    classifiers = {}
    for c in range(k):
        Ybin = (Y == c)
        result = train_fun(X, Ybin)
        classifiers[c] = result[:-1]
        loss = loss + result[-1]
    return classifiers, loss / k


def one_vs_rest_inference(X, inference_fun, params):
    """Multiclass prediction of the class labels.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    inference_fun : function that, given X and a tuple of parameters,
        returns prediction scores.
    params : dict
         parameters of the binary classifiers.

    Returns
    -------
    ndarray, shape (m,)
        predicted labels (one per feature vector) in the range 0...(k-1).
    ndarray, shape (m, k)
        classification scores.
    """
    k = len(params)
    scores = np.empty((X.shape[0], k))
    for c, p in params.items():
        scores[:, c] = inference_fun(X, *p)
    labels = np.argmax(scores, 1)
    return labels, scores
#!end1


#!begin2
def one_vs_one_train(X, Y, train_fun):
    """Train a multi-class classifier using the one vs. one strategy.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features.
    Y : ndarray, shape (m,)
        training labels in the range 0, ..., (k - 1).
    train_fun : function that, given X and a set of binary labels, trains
        a binary classifier.  The function must return a tuple, where the
        last element is the list of loss values obtained during training,
        and all the others are the parameters of the trained classifier.

    Returns
    -------
    params : dict
        collection of ((pos, neg), parameters) representing the binary
        classifiers.
    loss : ndarray, shape (steps,)
        average loss value after each training iteration.
    """
    k = Y.max() + 1
    loss = 0
    params = {}
    # For each pair of classes...
    for pos in range(k):
        for neg in range(pos + 1, k):
            # Build a training subset
            subset = (np.logical_or(Y == pos, Y == neg)).nonzero()[0]
            Xbin = X[subset, :]
            Ybin = (Y[subset] == pos)
            # Train the classifier
            result = train_fun(Xbin, Ybin)
            params[(pos, neg)] = result[:-1]
            loss = loss + result[-1]
    return params, loss / len(params)


def one_vs_one_inference(X, inference_fun, params):
    """Multiclass prediction of the class labels.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    inference_fun : function that, given X and a tuple of parameters,
        returns binary labels.
    params : dict
         parameters of the binary classifiers.

    Returns
    -------
    ndarray, shape (m,)
        predicted labels (one per feature vector) in the range 0...(k-1).
    ndarray, shape (m, k)
        classification scores.
    """
    m = X.shape[0]
    k = 1 + max(max(classes) for classes in params)
    votes = np.zeros((m, k))
    for classes, p in params.items():
        pos, neg = classes
        Y = inference_fun(X, *p)
        votes[Y > 0, pos] += 1
        votes[Y <= 0, neg] += 1
    labels = np.argmax(votes, 1)
    return labels, votes
#!end2


if __name__ == "__main__":
    import demo
    import ksvm

    class Demo(demo.Demo):
        def binary_train(self, X, Y):
            alpha, b, loss = ksvm.ksvm_train(X, Y, self.args.kernel,
                                             self.args.kparam,
                                             self.args.lambda_,
                                             lr=self.args.lr,
                                             steps=self.args.steps)
            return (X, alpha, b, loss)

        def ova_inference(self, X, Xtrain, alpha, b):
            return ksvm.ksvm_inference(X, Xtrain, alpha, b,
                                       self.args.kernel,
                                       self.args.kparam)[1]

        def ovo_inference(self, X, Xtrain, alpha, b):
            return ksvm.ksvm_inference(X, Xtrain, alpha, b,
                                       self.args.kernel,
                                       self.args.kparam)[0]

        def train(self, X, Y):
            if self.args.strategy == "ova":
                self.params, loss = one_vs_rest_train(X, Y, self.binary_train)
            else:
                self.params, loss = one_vs_one_train(X, Y, self.binary_train)
            return loss

        def inference(self, X):
            if self.args.strategy == "ova":
                return one_vs_rest_inference(X, self.ova_inference,
                                             self.params)[0]
            else:
                return one_vs_one_inference(X, self.ovo_inference,
                                            self.params)[0]

    app = Demo()
    app.parser.add_argument("--strategy", choices=["ova", "ovo"],
                            default="ova", help="multiclass strategy")
    app.parser.add_argument("-k", "--kernel", choices=["rbf", "polynomial"],
                            default="rbf", help="Kernel function")
    app.parser.add_argument("-g", "--kparam", type=float, default=3,
                            help="Parameter of the kernel")
    app.run()



import numpy as np


# TODO:
# - docstrings


def knn_inference(X, Xtrain, Ytrain, k=1):
    """K-Nearest Neighbors prediction of class labels.

    The function also estimates the posterior probabilities as
    fraction of neighbors voting for each class.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    Xtrain : ndarray, shape (t, n)
         reference (training) features (one row per feature vector).
    Ytrain : ndarray, shape (t,)
         reference (training) labels (integers in the range 0...(c - 1) ).
    k : int
         number of neighbors.

    Returns
    -------
    ndarray, shape (m,)
        predicted labels (one per feature vector).
    ndarray, shape (m, c)
        probability estimates (one per feature vector).

    """
    m = X.shape[0]
    classes = Ytrain.max() + 1
    D = _dist_matrix(X, Xtrain)
    if k == 1:
        index = np.argmin(D, 1)
        labels = Ytrain[index]
        probs = np.zeros((m, classes))
        probs[np.arange(m), labels] = 1
    else:
        neighs = np.argpartition(D, k, 1)[:, :k]
        counts = _bincount_rows(Ytrain[neighs], classes)
        labels = np.argmax(counts, 1)
        probs = counts / k
    return labels, probs


def knn_select_k(X, Y, maxk=101):
    """Leave-one-out selection of the number of neighbors for KNN.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         reference (training) features (one row per feature vector).
    Y : ndarray, shape (m,)
         reference (training) labels (integers in the range 0...(c - 1) ).
    maxk : int
         maximum value of k that is going to be evaluated.

    Returns
    -------
    int
        best value found for k.
    float
        accuracy estimated for the best k.
    """
    D = _dist_matrix(X, X)
    m = X.shape[0]
    classes = Y.max() + 1
    np.fill_diagonal(D, np.inf)
    neighs = np.argsort(D, 1)
    best_k = 1
    best_acc = -1
    for k in range(1, min(m, maxk + 1), 2):
        counts = _bincount_rows(Y[neighs[:, :k]], classes)
        labels = np.argmax(counts, 1)
        accuracy = (labels == Y).mean()
        if accuracy > best_acc:
            best_acc = accuracy
            best_k = k
    return best_k, best_acc


def _bincount_rows(X, values):
    """Compute one histogram per row.

    np.bincount works only on 1D arrays.  This extends it to work on
    each row.
    """
    m = X.shape[0]
    idx = X.astype(int) + values * np.arange(m)[:, np.newaxis]
    c = np.bincount(idx.ravel(), minlength=values * m)
    return c.reshape(-1, values)


def _dist_matrix(X1, X2):
    """Compute the matrix D.

    D[i, j] is the square of the distance between X1[i, :] and X2[j ,:].
    """
    Q1 = (X1 ** 2).sum(1, keepdims=True)
    Q2 = (X2 ** 2).sum(1, keepdims=True)
    return Q1 - 2 * X1 @ X2.T + Q2.T



import numpy as np


# TODO:
# - docstrings & comments
# - CC Pruning
# - categorical variables and splits


class Node:
    def inference(self, X, classes):
        pass

    def _dump(self, indent):
        pass


class TerminalNode(Node):
    def __init__(self, distribution):
        self.distribution = distribution

    def inference(self, X, classes):
        return np.tile(self.distribution, (X.shape[0], 1))

    def _dump(self, indent):
        s = np.array_str(self.distribution, precision=4)
        p = self.distribution.argmax()
        return (" " * indent) + s + " => " + str(p) + "\n"


class InternalNode(Node):
    def __init__(self, left, right, feature, value):
        self.left = left
        self.right = right
        self.feature = feature
        self.value = value

    def inference(self, X, classes):
        out = np.empty((X.shape[0], classes))
        ii = (X[:, self.feature] < self.value)
        if ii.any():
            out[ii, :] = self.left.inference(X[ii, :], classes)
        if not ii.all():
            ii = np.logical_not(ii)
            out[ii, :] = self.right.inference(X[ii, :], classes)
        return out

    def _dump(self, indent):
        i = (" " * indent)
        hd = i + "if x[{}] < {}:\n".format(self.feature, self.value)
        lb = self.left._dump(indent + 4)
        el = i + "else:\n"
        rb = self.right._dump(indent + 4)
        return "".join([hd, lb, el, rb])


class ClassificationTree:
    def __init__(self):
        self.root = TerminalNode(0)
        self.classes = 1

    def inference(self, X):
        probs = self.root.inference(X, self.classes)
        labels = probs.argmax(1)
        return labels, probs

    def train(self, X, Y, diversity="gini", minsize=10, classes=None):
        dfun = _DIVERSITY_FUN[diversity]
        if classes is None:
            self.classes = Y.max() + 1
        m = X.shape[0]
        H = np.zeros((m, self.classes), dtype=int)
        H[np.arange(m), Y] = 1

        def _train_rec(X, H):
            if X.shape[0] < minsize:
                dist = (1 + H.sum(0)) / (X.shape[0] + self.classes)
                return TerminalNode(dist)
            split = _find_split(X, H, dfun, minsize)
            if split is None:
                dist = (1 + H.sum(0)) / (X.shape[0] + self.classes)
                return TerminalNode(dist)
            ii = (X[:, split[0]] < split[1])
            left = _train_rec(X[ii, :], H[ii, :])
            ii = np.logical_not(ii)
            right = _train_rec(X[ii, :], H[ii, :])
            return InternalNode(left, right, split[0], split[1])
        self.root = _train_rec(X, H)


def _find_split(X, H, criterion, minsize):
    m, n = X.shape
    if m < 2 * minsize:
        return None
    split = None
    for j in range(n):
        ret = _decision_stump(X[:, j], H, criterion, minsize)
        if ret is not None and (split is None or ret[1] < split[2]):
            split = j, ret[0], ret[1]
    return split


_DIVERSITY_FUN = {
    "gini": lambda p: (1 - (p ** 2).sum(1)),
    "entropy": lambda p: (-(p * np.nan_to_num(np.log(p))).sum(1)),
    "error": lambda p:  (1 - p.max(1))
}


def _decision_stump(x, h, criterion, minsize):
    m, k = h.shape
    # 1) sort the data and find candidate splits
    ii = np.argsort(x)
    # possible split points are those where consecutive values are
    # different, except for the first and last minsize positions.
    sp = (x[ii[minsize - 1:m - minsize]] < x[ii[minsize:m - minsize + 1]])
    sp = sp.nonzero()[0] + minsize - 1
    if sp.size == 0:
        return None
    # 2) compute the class distributions at the split points
    counts_lo = h[ii, :].cumsum(0)
    counts_hi = counts_lo[-1:, :] - counts_lo
    p_lo = ((1 + counts_lo[sp, :]) /
            (k + counts_lo[sp, :].sum(1, keepdims=True)))
    p_hi = ((1 + counts_hi[sp, :]) /
            (k + counts_hi[sp, :].sum(1, keepdims=True)))
    p_split = sp / m
    # 3) compute the average criterion and select the lowest
    cost = p_split * criterion(p_lo) + (1 - p_split) * criterion(p_hi)
    j = cost.argmin()
    # 4) return the split value (midway the two consecutvie samples)
    # and the corresponding cost
    best = sp[j]
    threshold = (x[ii[best + 1]] + x[ii[best]]) / 2
    return (threshold, cost[j])



import ksvm


def grid_search_ksvm_rbf(Xtrain, Ytrain, Xval, Yval, kfun, lambdas,
                         kparams, lr=1e-3, steps=1000):
    best_accuracy = 0
    best_alpha = None
    best_b = None
    best_kparam = None
    results = []
    for lambda_ in lambdas:
        results.append([])
        for kparam in kparams:
            alpha, b, loss = ksvm.ksvm_train(Xtrain, Ytrain, kfun,
                                             kparam, lambda_, lr=lr,
                                             steps=steps)
            T = ksvm.ksvm_inference(Xval, Xtrain, alpha, b, kfun,
                                    kparam)
            acc = accuracy(Yval, T)
            results[-1].append(acc)
            if acc > best_accuracy:
                best_accuracy = acc
                best_lambda_ = lambda_
                best_kparam = kparam
                best_alpha = alpha
                best_b = b
    return (best_accuracy, best_lambda_, best_kparam, best_alpha,
            best_b, results)


def accuracy(Y1, Y2):
    return (Y1 == Y2).mean()


def _demo():
    import sys
    import demo
    lambdas = [0.001, 0.003, 0.01, 0.03, 0.1]
    gammas = [0.1, 0.3, 1, 3, 10, 30]
    Xtrain, Ytrain = demo.load_data(sys.argv[1], True)
    Xval, Yval = demo.load_data(sys.argv[2], True)
    res = grid_search_ksvm_rbf(Xtrain, Ytrain, Xval, Yval, "rbf",
                               lambdas, gammas, steps=10000)
    acc, lambda_, gamma, alpha, b, results = res
    for row in results:
        for col in row:
            print("%.4f" % col, end=" ")
        print()
    print("Best validation accuracy:", acc)
    print(alpha)
    print(b)


if __name__ == "__main__":
    _demo()



import numpy as np


def hgda_train(X, Y, priors=None):
    """Train a heteroscedastic GDA classifier.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features.
    Y : ndarray, shape (m,)
        training labels with values in {0, ..., k - 1}.
    priors : ndarray, shape (k,)
        Prior probabilities for the classes (if None they get
        estimated from Y).

    Returns
    -------
    means : ndarray, shape (k, n)
         class mean vectors.
    invcovs : ndarray, shape (k, n, n)
         inverse of the class covariance matrices.
    priors : ndarray, shape (k,)
         class prior probabilities.
    """
    k = Y.max() + 1
    m, n = X.shape
    means = np.empty((k, n))
    invcovs = np.empty((k, n, n))
    if priors is None:
        priors = np.bincount(Y) / m
    for c in range(k):
        indices = (Y == c).nonzero()[0]
        means[c, :] = X[indices, :].mean(0)
        cov = np.cov(X[indices, :].T)
        invcovs[c, :, :] = np.linalg.inv(cov)
    return (means, invcovs, priors)


def hgda_inference(X, means, invcovs, priors):
    """Heteroscedastic GDA inference.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    means : ndarray, shape (k, n)
         class mean vectors.
    invcovs : ndarray, shape (k, n, n)
         inverse of the class covariance matrices.
    priors : ndarray, shape (k,)
         class prior probabilities.

    Returns
    -------
    ndarray, shape (m,)
        predicted labels (one per feature vector).
    ndarray, shape (m, k)
        scores assigned to each class.
    """
    m, n = X.shape
    k = means.shape[0]
    scores = np.empty((m, k))
    for c in range(k):
        det = np.linalg.det(invcovs[c, :, :])
        diff = X - means[None, c, :]
        q = ((diff @ invcovs[c, :, :]) * diff).sum(1)
        scores[:, c] = 0.5 * q - 0.5 * np.log(det) - np.log(priors[c])
    labels = np.argmin(scores, 1)
    return labels, -scores


def ogda_train(X, Y, priors=None):
    """Train a omoscedastic GDA classifier.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features.
    Y : ndarray, shape (m,)
        training labels with values in {0, ..., k - 1}.
    priors : ndarray, shape (k,)
        Prior probabilities for the classes (if None they get
        estimated from Y).

    Returns
    -------
    W : ndarray, shape (n, k)
         weight vectors, each row representing a different class.
    b : ndarray, shape (k,)
         vector of biases.
    """
    k = Y.max() + 1
    m, n = X.shape
    means = np.empty((k, n))
    cov = np.zeros((n, n))
    if priors is None:
        priors = np.bincount(Y) / m
    for c in range(k):
        indices = (Y == c).nonzero()[0]
        means[c, :] = X[indices, :].mean(0)
        cov += priors[c] * np.cov(X[indices, :].T)
    icov = np.linalg.inv(cov)
    W = -(icov @ means.T)
    q = ((means @ icov) * means).sum(1)
    b = 0.5 * q - np.log(priors)
    return (W, b)


def ogda_inference(X, W, b):
    """Omoscedastic GDA inference.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    W : ndarray, shape (n, k)
         weight vectors, each row representing a different class.
    b : ndarray, shape (k,)
         vector of biases.

    Returns
    -------
    ndarray, shape (m,)
        predicted labels (one per feature vector).
    ndarray, shape (m, k)
        scores assigned to each class.
    """
    scores = X @ W + b.T
    labels = np.argmin(scores, 1)
    return labels, -scores


def mindist_train(X, Y):
    """Train a minimum distance classifier.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features.
    Y : ndarray, shape (m,)
        training labels with values in {0, ..., k - 1}.

    Returns
    -------
    means : ndarray, shape (k, n)
         class mean vectors.
    """
    k = Y.max() + 1
    n = X.shape[1]
    means = np.empty((k, n))
    for c in range(k):
        indices = (Y == c).nonzero()[0]
        means[c, :] = X[indices, :].mean(0)
    return (means)


def mindist_inference(X, means):
    """Minimum distance classifier inference.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    means : ndarray, shape (k, n)
         class mean vectors.

    Returns
    -------
    ndarray, shape (m,)
        predicted labels (one per feature vector).
    ndarray, shape (m, k)
        scores assigned to each class.
    """
    sqdists = ((X[:, None, :] - means[None, :, :]) ** 2).sum(2)
    labels = np.argmin(sqdists, 1)
    return labels, -sqdists



import numpy as np
import matplotlib.pyplot as plt
import scipy.interpolate


def circle_data(n, k):
    t = np.linspace(0, 2 * np.pi, n)
    X = np.vstack([np.cos(t), np.sin(t)]).T
    Y = np.repeat(np.arange(k), (n + k - 1) // k)[:n]
    return X, Y


def plot_linear(w, b, xmin, xmax, ymin, ymax, *args, **kwargs):
    if np.abs(w[1]) > np.abs(w[0]):
        y1 = -(b + w[0] * xmin) / w[1]
        y2 = -(b + w[0] * xmax) / w[1]
        plt.plot([xmin, xmax], [y1, y2], *args, **kwargs)
    else:
        x1 = -(b + w[1] * ymin) / w[0]
        x2 = -(b + w[1] * ymax) / w[0]
        plt.plot([x1, x2], [ymin, ymax], *args, **kwargs)


def apply_grid(f, fromx, tox, fromy, toy, resolution=100):
    ax = np.linspace(fromx, tox, resolution)
    ay = np.linspace(fromy, toy, resolution)
    x, y = np.meshgrid(ax, ay)
    data = np.vstack((x.reshape(-1), y.reshape(-1))).T
    z = f(data).reshape(x.shape)
    return (x, y, z)


def contour_fun(f, values, fromx, tox, fromy, toy, resolution=100, **opts):
    ax = np.linspace(fromx, tox, resolution)
    ay = np.linspace(fromy, toy, resolution)
    x, y = np.meshgrid(ax, ay)
    data = np.vstack((x.reshape(-1), y.reshape(-1))).T
    v = f(data).reshape(x.shape)
    return plt.contour(x, y, v, values, **opts)


def image_fun(f, fromx, tox, fromy, toy, resolution=100, **opts):
    ax = np.linspace(fromx, tox, resolution)
    ay = np.linspace(fromy, toy, resolution)
    x, y = np.meshgrid(ax, ay)
    data = np.vstack((x.reshape(-1), y.reshape(-1))).T
    v = f(data).reshape(x.shape)
    # return plt.contour(x, y, v, values, **opts)
    return plt.imshow(v, **opts)


def check_gradient(x, f, gf, eps=1e-5, toll=1e-4):
    g = gf(x)
    x1 = x + eps * g
    x2 = x - eps * g
    y1 = f(x1)
    y2 = f(x2)
    d = y1 - y2 - 2 * eps * (g ** 2).sum()
    if np.abs(d) >= toll:
        print(x)
        print(g)
        print(d)
    assert np.abs(d) < toll, "Wrong gradient approximation: %f" % d


def check_gradient2(x, f, gf, eps=1e-5):
    n = x.shape[0]
    d = np.empty_like(x)
    for i in range(n):
        x1 = x.copy()
        x1[i] += eps
        x2 = x.copy()
        x2[i] -= eps
        y1 = f(x1)
        y2 = f(x2)
        d[i] = y1 - y2 - 2 * eps
    print(d)
    assert np.abs(d) < 1000 * eps


def save_contour(c, filename, key=0, fmt="%f", k=5):
    """Save a matplotlib contour in a text file."""
    segs = c.collections[key].get_segments()
    m = max(s.shape[0] for s in segs)
    data = np.nan * np.ones((k * m - k + 1, 2 * len(segs)))
    for n, s in enumerate(segs):
        nx = s.shape[0]
        kx = k * nx - k + 1
        for i in (0, 1):
            ius = scipy.interpolate.InterpolatedUnivariateSpline(
                np.arange(nx), s[:, i])
            data[:kx, 2 * n + i] = ius(np.linspace(0, nx - 1, kx))
    np.savetxt(filename, data, fmt=fmt)



import numpy as np
from .mlp import relu, softmax


class CNN:
    """Convolutional neural network.

    It is composed by a sequence of convolutional layers, followed by
    an average pooling and a softmax.

    The activation function is relu (except for the last convolution).

    The loss function used during training is the cross entropy.

    The network processes four dimensional arrays representing batch
    of images.  The four dimensions represents (1) batch size (2)
    height (3) width (4) channels.

    """

    def __init__(self, channels, kernel_sz=None, strides=None):
        """Create and initialiaze the CNN.

        Parameters
        ----------
        channels : list
            number of channels in the layers (first input, then hiddens,
            then output).  At least two are required.
        kernel_sz : list
            size (height and width) of the spatial filters used by
            convolutions.  Must contain one size for each layer,
            except input.  If None, all filters are 3x3.

        strides : list
            stride (vertical and horizontal) used to apply filters.
            Must contain one for each layer, except input.  If None
            all strides are one.

        For instance, CNN([3, 16, 10], [5, 3], [2, 1]) creates a CNN
        with two convolutional layers: the first is a 5x5 convolution
        with stride 2, 3 input and 16 output channels.  The second is
        a 3x3 convolution with stride 1, 16 input and 10 output
        channels.

        """
        # Defaults
        kernel_sz = (kernel_sz or [3] * (len(channels) - 1))
        self.strides = (strides or [1] * (len(channels) - 1))
        # Initialize weights with the Kaiming technique
        self.weights = [
            np.random.randn(k, k, m, n) * np.sqrt(2.0 / (k * k * m))
            for m, n, k in zip(channels[:-1], channels[1:], kernel_sz)
        ]
        # Biases are zero-initialized
        self.biases = [np.zeros(m) for m in channels[1:]]
        # Accumulators for the momentum terms
        self.update_w = [np.zeros_like(w) for w in self.weights]
        self.update_b = [np.zeros_like(b) for b in self.biases]

    def forward(self, X):
        """Compute the activations of the neurons.

        Parameters
        ----------
        X : ndarray, shape (m, h, w, n)
            input images (m images of n channels).

        Returns
        -------
        list
            the list of activations of the neurons

        The returned list contains arrays (one per layer) with the
        activations of the layers.  The first element of the list is
        the input X, followed by the activations of hidden layers and
        terminated by the activations in the output layer.

        """
        activations = [X]
        for W, b, s in zip(self.weights, self.biases, self.strides):
            X = _convolution(X, W, s, s) + b
            if W is not self.weights[-1]:
                X = relu(X)
            else:
                activations.append(X)
                X = X.mean(1).mean(1)
                X = softmax(X)
            activations.append(X)
        return activations

    def backward(self, Y, activations):
        """Compute the derivatives of the loss wrt the activations.

        Parameters
        ----------
        Y : ndarray, shape (m,)
            target output (integer class labels).
        activations : list
            activations computed by the forward method.

        Returns
        -------
        list
            the list of computed derivatives, one for each layer.
        """
        d = activations[-1].copy()
        d[np.arange(Y.shape[0]), Y] -= 1  # Subtract the one-hot vectors
        sz = activations[-2].shape[1] * activations[-2].shape[2]
        d /= (sz * activations[-1].shape[0])
        d = d[:, None, None, :].repeat(activations[-2].shape[1], 1)
        d = d.repeat(activations[-2].shape[2], 2)
        derivatives = [d]
        for W, X, s in zip(self.weights[:0:-1], activations[-3::-1],
                           self.strides[:0:-1]):
            d = _convolution_backprop(d, W, X.shape[1], X.shape[2], s, s)
            derivatives.append(d)
            d *= (X > 0).astype(int)  # derivative of relu
        return derivatives[::-1]

    def loss(self, Y, P):
        """Compute the average cross-entropy."""
        return -np.log(P[np.arange(Y.shape[0]), Y]).mean()

    def backpropagation(self, X, Y, lr=1e-4, lambda_=1e-5, momentum=0.99):
        """Backpropagation algorithm.

        Perform both the forward and the backward steps and update the
        parameters.

        Parameters
        ----------
        X : ndarray, shape (m, h, w, n)
            input images (m images of n channels).
        Y : ndarray, shape (m,)
            target output (integer class labels).
        lr : float
            learning rate.
        lambda_ : float
            regularization coefficients.
        momentum : float
            momentum coefficient.

        """
        activations = self.forward(X)
        derivatives = self.backward(Y, activations)
        for X, D, W, b, s, uw, ub in zip(activations, derivatives,
                                         self.weights, self.biases,
                                         self.strides,
                                         self.update_w, self.update_b):
            grad_b = D.sum(2).sum(1).sum(0)
            grad_W = _convolution_derivative(X, D, W.shape[0],
                                             W.shape[1], s, s)
            grad_W += lambda_ * W
            uw *= momentum
            uw -= lr * grad_W
            W += uw
            ub *= momentum
            ub -= lr * grad_b
            b += ub

    def inference(self, X):
        """Compute the predictions of the network.

        Parameters
        ----------
        X : ndarray, shape (m, h, w, n)
            input images (m images of n channels).

        Returns
        -------
        ndarray, shape (m,)
            predicted labels, in the range 0, ..., k - 1
        ndarray, shape (m, k)
            posterior probability estimates.

        """
        probs = self.forward(X)[-1]
        labels = np.argmax(probs, 1)
        return labels, probs

    def train(self, X, Y, lr=1e-4, lambda_=1e-5, momentum=0.99,
              steps=10000, batch=None):
        """Train the network.

        Apply multiple steps of stochastic gradient descent.

        Parameters
        ----------
        X : ndarray, shape (m, h, w, n)
            input images (m images of n channels).
        Y : ndarray, shape (m,)
            target output (integer class labels).
        lr : float
            learning rate.
        lambda_ : float
            regularization coefficients.
        momentum : float
            momentum coefficient.
        steps : int
            training iterations.
        batch : int or None
            size of the minibatch used in each step.  When None all
            the data is used in each step.

        """
        m = X.shape[0]
        if batch is None:
            batch = X.shape[0]
        i = m
        indices = np.arange(m)
        for step in range(steps):
            if i + batch > m:
                i = 0
                np.random.shuffle(indices)
            self.backpropagation(X[indices[i:i + batch], :, :, :],
                                 Y[indices[i:i + batch]], lr=lr,
                                 lambda_=lambda_, momentum=momentum)
            i += batch

    def save(self, filename):
        """Save the network to the file."""
        np.savez(filename, weights=self.weights, biases=self.biases,
                 strides=self.strides)

    @classmethod
    def load(cls, filename):
        """Create a new network from the data saved in the file."""
        data = np.load(filename, allow_pickle=True)
        neurons = [w.shape[0] for w in data["weights"]]
        neurons.append(data["weights"][-1].shape[1])
        network = cls(neurons)
        network.weights = data["weights"]
        network.biases = data["biases"]
        network.strides = data["strides"]
        return network


# The convolution operator and its derivatives are implemented as
# described in the blog post:
#
# https://medium.com/@mayank.utexas/backpropagation-for-convolution-with-strides-8137e4fc2710


def _im2col(X, kh, kw, sh, sw):
    """Transform an image into an array of neighborhoods.

    Rectangular neighborhoods of size (kw, kh, n) are linearized into
    a single dimension with size (kw * kh * n).

    Parameters
    ----------
    X : ndarray, shape (m, h, w, n)
        input images (m images of n channels).
    kh : int
        rows in each neighborhood.
    kw : int
        columns in each neighborhood.
    sh : int
        rows between two adjacent neighborhoods (stride).
    sw : int
        columns between two adjacent neighborhoods (stride).

    Returns
    -------
    ndarray, shape (m, nh, nw, nn)
        Array of neighborhoods (nh = (h - kh + sh) // sh,
        nw = (w - kw + sw) // sw, nn = n * kh * kw).

    """
    m, h, w, n = X.shape
    shape = (m, (h - kh + sh) // sh, (w - kw + sw) // sw, kh, kw, n)
    tm, th, tw, tn = X.strides
    strides = (tm, sh * th, sw * tw, th, tw, tn)
    Y = np.lib.stride_tricks.as_strided(X, shape=shape, strides=strides,
                                        writeable=False)
    return Y.reshape(m, shape[1], shape[2], kh * kw * n)


def _convolution(X, W, sh, sw):
    """Two-dimensional convolution of a batch of images.

    Parameters
    ----------
    X : ndarray, shape (m, h, w, n)
        input images (m images of n channels).
    W : ndarray, shape (kh, kw, n, c)
        filter coefficients.
    sh : int
        vertical stride.
    sw : int
        horizontal stride.

    Returns
    -------
    ndarray, shape (m, nh, nw, nn)
        Result of the convolution (nh = (h - kh + sh) // sh,
        nw = (w - kw + sw) // sw).

    Output[b, i, j, c] =
       sum[u,v,d]( X[b, sh * i + u, sw * j + v, d] * W[u, v, d, c] )
    """
    assert X.ndim == 4
    assert X.shape[3] == W.shape[2]
    assert X.shape[1] >= W.shape[0]
    assert X.shape[2] >= W.shape[1]
    Y = _im2col(X, W.shape[0], W.shape[1], sh, sw)
    Z = Y.reshape(-1, Y.shape[3]) @ W.reshape(-1, W.shape[3])
    return Z.reshape(Y.shape[0], Y.shape[1], Y.shape[2], -1)


def _dilate(X, h, w, ph, pw, sh, sw):
    """Dilation operator.

    Insert gaps (filled with zeros) between the elments of X.

    Parameters
    ----------
    X : ndarray, shape (m, h, w, n)
        input images (m images of n channels).
    h : int
        height of the result.
    w : int
        width of the result.
    ph : int
        row where the top elements of X are placed.
    pw : int
        column where the leftmost elements of X are placed.
    sh : int
        vertical stride.
    sw : int
        horizontal stride.

    Returns
    -------
    ndarray, shape (m, h, w, n)

    """
    #          00000
    # 12       01020
    # 34  == > 00000
    #          03040
    #          00000
    b, hin, win, c = X.shape
    Y = np.zeros((b, h, w, c))
    Y[:, ph:(ph + sh * hin):sh, pw:(pw + sw * win):sw, :] = X
    return Y


def _convolution_backprop(D, W, h, w, sh, sw):
    """Return the derivative of the convolution.

    Parameters
    ----------
    D : ndarray, shape (m, hd, wd, c)
        derivative of the result of the convolution
    W : ndarray, shape (kh, kw, n, c)
        filter coefficients.
    h : int
        height of the argument of the convolution
    w : int
        width of the argument of the convolution
    sh : int
        vertical stride.
    sw : int
        horizontal stride.

    Returns
    -------
    ndarray, shape (m, h, w, n)
        derivative of the convolution with respect to its argument.

    h and w are required, since the size of the argument cannot be
    inferred when the stride is greater than one.
    """
    # Dilate and pad the derivative
    kh = W.shape[0]
    kw = W.shape[1]
    D = _dilate(D, h + kh - 1, w + kw - 1, kh - 1, kw - 1, sh, sw)
    # The coefficient of the filters are mirrored and transposed
    W1 = W[::-1, ::-1, :, :].transpose(0, 1, 3, 2)
    return _convolution(D, W1, 1, 1)


def _convolution_derivative(X, D, kh, kw, sh, sw):
    """Derivative with respect to the filter coefficients.

    Parameters
    ----------
    X : ndarray, shape (m, h, w, n)
        input images (m images of n channels).
    D : ndarray, shape (m, hd, wd, c)
        derivative of the result of the convolution
    kh : int
        height of the convolutional kernel
    kw : int
        width of the convolutional kernel
    sh : int
        vertical stride
    sw : int
        horizontal stride

    Returns
    -------
    ndarray, shape (kh, kw, n, c)
        derivatives with respect to the filter coefficients
    """
    h = 1 + sh * (D.shape[1] - 1)
    w = 1 + sw * (D.shape[2] - 1)
    D = _dilate(D, h, w, 0, 0, sh, sw)
    D = D.transpose(1, 2, 0, 3)
    X = X.transpose(3, 1, 2, 0)
    X = X[:, :h + kh - 1, :w + kw - 1, :]
    R = _convolution(X, D, 1, 1)
    return R.transpose(1, 2, 0, 3)



import numpy as np
import urllib.request
import gzip
import sys


_FACTORIES = {}


def load_dataset(datasetname):
    """Load or generate a data set given its name.

    If the name is unknown then it interprets it as a path to a data
    file.  The file should be a text file with one vector per row and
    with elements separated by spaces).  The last element is used as
    class label.

    Parameters
    ----------
    datasetname : str
         name of the dataset or path to a file

    Returns
    -------
    ndarray, shape (m, n)
        features.
    ndarray, shape (m,)
        class labels.

    """
    try:
        f = _FACTORIES[datasetname]
        return f()
    except KeyError:
        pass
    data = np.loadtxt(datasetname)
    X = data[:, :-1]
    Y = data[:, -1].astype(int)
    return X, Y


def _register(name):
    def wrapper(f):
        _FACTORIES[name] = f
        return f
    return wrapper


@_register("gaussian")
def _gaussian_dataset():
    """Two Gaussians."""
    sz = 100
    Y = np.arange(sz) % 2
    X = np.random.randn(sz, 2)
    X[:, 0] += 3 * Y
    return X, Y


@_register("gaussian4")
def _gaussian4_dataset():
    """Four gaussians."""
    sz = 100
    Y = np.arange(sz) % 4
    X = np.random.randn(sz, 2)
    X[:, 0] += 4 * (Y // 2)
    X[:, 1] += 4 * (Y % 2)
    return X, Y


@_register("xor")
def _xor_dataset():
    """Exclusive OR configuration."""
    sz = 100
    X = np.random.randn(sz, 2) / 10
    X[(sz // 2):, 0] += 1
    X[(sz // 4):(3 * sz // 4), 1] += 1
    Y = np.logical_xor(X[:, 0] > 0.5, X[:, 1] > 0.5).astype(int)
    return X, Y


@_register("rings")
def _rings_dataset():
    """Two concentric rings."""
    sz = 100
    Y = np.arange(sz) % 2
    t = 2 * np.pi * np.random.rand(sz)
    r = Y + 1
    X = np.vstack([r * np.cos(t), r * np.sin(t)]).T
    X += 0.1 * np.random.randn(*X.shape)
    return X, Y


@_register("moons")
def _moons_dataset():
    """Two non-linearly separable moons."""
    sz = 100
    Y = np.arange(sz) % 2
    t = np.linspace(0, np.pi, sz)
    X = np.vstack([np.cos(t + Y * np.pi), np.sin(t + Y * np.pi)]).T
    X += np.outer(Y, np.array([1, 0.3])) + 0.1 * np.random.randn(*X.shape)
    return X, Y


@_register("swissroll")
def _swissroll_dataset():
    """Interwined spirals."""
    sz = 100
    Y = np.arange(sz) % 2
    t = np.linspace(0, 4 * np.pi, sz)
    X = t[:, None] * np.vstack([np.cos(t + Y * np.pi),
                                np.sin(t + Y * np.pi)]).T
    X += 0.2 * np.random.randn(*X.shape)
    return X, Y


@_register("categorical")
def _categorical_dataset():
    """Discrete features."""
    sz = 100
    t = np.arange(sz) % 36
    X = np.vstack([t % 6, t // 6]).T
    Y = ((t % 3) == 2).astype(int)
    s = (np.random.rand(sz) > 0.9)
    Y[s] = 1 - Y[s]
    X = X + np.random.random(X.shape) * 0.05
    return X, Y


@_register("yinyang")
def _yinyang_dataset():
    """Yin and yang shape."""
    sz = 200
    t = 2 * np.pi * np.random.rand(sz)
    r = 6 * (np.random.rand(sz) ** 0.5)
    X = np.vstack([r * np.cos(t), r * np.sin(t)]).T
    r1 = ((X - np.array([[0, 3]])) ** 2).sum(1)
    r2 = ((X - np.array([[0, -3]])) ** 2).sum(1)

    Y = np.logical_and(np.logical_or(r1 < 9, X[:, 0] > 0),
                       np.logical_not(r2 < 9))
    Y2 = np.logical_or(r1 < 1.5 ** 2, r2 < 1.5 ** 2)
    Y = np.logical_xor(Y, Y2).astype(int)
    return X, Y


@_register("iris")
def _iris_dataset():
    """The famous 'iris' dataset.

    This version provides only two features: the sepal length and the
    sepal width.

    The three classes are:
      0 => Iris setosa
      1 => Iris versicolor
      2 => Iris virginica

    """
    data = _IRIS_TXT.split()
    x1 = list(map(float, data[0::3]))
    x2 = list(map(float, data[1::3]))
    X = np.stack((np.array(x1), np.array(x2)), 1)
    Y = np.array(list(map(int, data[2::3])))
    return X, Y


_IRIS_TXT = """5.1 3.5 0 4.9 3 0 4.7 3.2 0 4.6 3.1 0 5 3.6 0 5.4 3.9 0 4.6 3.4 0 5
3.4 0 4.4 2.9 0 4.9 3.1 0 5.4 3.7 0 4.8 3.4 0 4.8 3 0 4.3 3 0 5.8 4 0
5.7 4.4 0 5.4 3.9 0 5.1 3.5 0 5.7 3.8 0 5.1 3.8 0 5.4 3.4 0 5.1 3.7 0
4.6 3.6 0 5.1 3.3 0 4.8 3.4 0 5 3 0 5 3.4 0 5.2 3.5 0 5.2 3.4 0 4.7
3.2 0 4.8 3.1 0 5.4 3.4 0 5.2 4.1 0 5.5 4.2 0 4.9 3.1 0 5 3.2 0 5.5
3.5 0 4.9 3.1 0 4.4 3 0 5.1 3.4 0 5 3.5 0 4.5 2.3 0 4.4 3.2 0 5 3.5 0
5.1 3.8 0 4.8 3 0 5.1 3.8 0 4.6 3.2 0 5.3 3.7 0 5 3.3 0 7 3.2 1 6.4
3.2 1 6.9 3.1 1 5.5 2.3 1 6.5 2.8 1 5.7 2.8 1 6.3 3.3 1 4.9 2.4 1 6.6
2.9 1 5.2 2.7 1 5 2 1 5.9 3 1 6 2.2 1 6.1 2.9 1 5.6 2.9 1 6.7 3.1 1
5.6 3 1 5.8 2.7 1 6.2 2.2 1 5.6 2.5 1 5.9 3.2 1 6.1 2.8 1 6.3 2.5 1
6.1 2.8 1 6.4 2.9 1 6.6 3 1 6.8 2.8 1 6.7 3 1 6 2.9 1 5.7 2.6 1 5.5
2.4 1 5.5 2.4 1 5.8 2.7 1 6 2.7 1 5.4 3 1 6 3.4 1 6.7 3.1 1 6.3 2.3 1
5.6 3 1 5.5 2.5 1 5.5 2.6 1 6.1 3 1 5.8 2.6 1 5 2.3 1 5.6 2.7 1 5.7 3
1 5.7 2.9 1 6.2 2.9 1 5.1 2.5 1 5.7 2.8 1 6.3 3.3 2 5.8 2.7 2 7.1 3 2
6.3 2.9 2 6.5 3 2 7.6 3 2 4.9 2.5 2 7.3 2.9 2 6.7 2.5 2 7.2 3.6 2 6.5
3.2 2 6.4 2.7 2 6.8 3 2 5.7 2.5 2 5.8 2.8 2 6.4 3.2 2 6.5 3 2 7.7 3.8
2 7.7 2.6 2 6 2.2 2 6.9 3.2 2 5.6 2.8 2 7.7 2.8 2 6.3 2.7 2 6.7 3.3 2
7.2 3.2 2 6.2 2.8 2 6.1 3 2 6.4 2.8 2 7.2 3 2 7.4 2.8 2 7.9 3.8 2 6.4
2.8 2 6.3 2.8 2 6.1 2.6 2 7.7 3 2 6.3 3.4 2 6.4 3.1 2 6 3 2 6.9 3.1 2
6.7 3.1 2 6.9 3.1 2 5.8 2.7 2 6.8 3.2 2 6.7 3.3 2 6.7 3 2 6.3 2.5 2
6.5 3 2 6.2 3.4 2 5.9 3 2 """


def _load_mnist_set(filename, features_url, labels_url):
    try:
        data = np.load(filename, allow_pickle=True)
        return data["X"], data["Y"]
    except FileNotFoundError:
        pass

    print("Downloading the data set...", file=sys.stderr)
    with urllib.request.urlopen(features_url) as remote:
        with gzip.open(remote) as f:
            data = np.frombuffer(f.read(), dtype=np.uint8)
            X = data[16:].reshape((-1, (28 * 28))).astype(np.float32) / 255.0
    with urllib.request.urlopen(labels_url) as remote:
        with gzip.open(remote) as f:
            data = np.frombuffer(f.read(), dtype=np.uint8)
            Y = data[8:].astype(np.int32)

    try:
        np.savez_compressed(filename, X=X, Y=Y)
        print("Dataset saved to '{}'".format(filename), file=sys.stderr)
    except Exception as e:
        print("Failed to save the data set to '{}' ({})".format(filename, e))
    return X, Y


_MNIST_URL = "http://yann.lecun.com/exdb/mnist/"


@_register("mnist_train")
def _mnist_train_dataset():
    """MNIST dataset for handwritten digits recognition (training data).

    The data is downloaded from http://yann.lecun.com/exdb/mnist/
    and saved to the file 'mnist_train.npz' to avoid multiple downloads.
    """
    X, Y = _load_mnist_set("mnist_train.npz",
                           _MNIST_URL + "train-images-idx3-ubyte.gz",
                           _MNIST_URL + "train-labels-idx1-ubyte.gz")
    return X, Y


@_register("mnist_test")
def _mnist_test_dataset():
    """MNIST dataset for handwritten digits recognition (test data).

    The data is downloaded from http://yann.lecun.com/exdb/mnist/
    and saved to the file 'mnist_test.npz' to avoid multiple downloads.
    """
    X, Y = _load_mnist_set("mnist_test.npz",
                           _MNIST_URL + "t10k-images-idx3-ubyte.gz",
                           _MNIST_URL + "t10k-labels-idx1-ubyte.gz")
    return X, Y


_FASHION_MNIST_URL = ("http://fashion-mnist.s3-website.eu-central-1" +
                      ".amazonaws.com/")


@_register("fashion_mnist_train")
def _fashion_mnist_train_dataset():
    """MNIST dataset for handwritten digits recognition (training data).

    The data is downloaded from http://yann.lecun.com/exdb/mnist/
    and saved to the file 'mnist_train.npz' to avoid multiple downloads.
    """
    X, Y = _load_mnist_set("fashion_mnist_train.npz",
                           _FASHION_MNIST_URL + "train-images-idx3-ubyte.gz",
                           _FASHION_MNIST_URL + "train-labels-idx1-ubyte.gz")
    return X, Y


@_register("fashion_mnist_test")
def _fashion_mnist_test_dataset():
    """MNIST dataset for handwritten digits recognition (test data).

    The data is downloaded from http://yann.lecun.com/exdb/mnist/
    and saved to the file 'mnist_test.npz' to avoid multiple downloads.
    """
    X, Y = _load_mnist_set("fashion_mnist_test.npz",
                           _FASHION_MNIST_URL + "t10k-images-idx3-ubyte.gz",
                           _FASHION_MNIST_URL + "t10k-labels-idx1-ubyte.gz")
    return X, Y


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    if len(sys.argv) < 2:
        print("No dataset given")
        print("Available datasets are:")
        for k, v in sorted(_FACTORIES.items()):
            txt = v.__doc__.splitlines()[0]
            print("{:20s}\t{}".format("'" + k + "'", txt))
        sys.exit()

    try:
        X, Y = load_dataset(sys.argv[1])
    except Exception as e:
        print("Error loading the data set ({})".format(e))
        sys.exit()
    m, n = X.shape
    print("Loaded {} samples with {} features".format(m, n))
    print()
    for k in range(Y.max() + 1):
        count = sum(Y == k)
        print("{} samples of class {}".format(count, k))
    if n == 2:
        plt.scatter(X[:, 0], X[:, 1], c=Y, cmap=plt.cm.Paired)
        plt.show()
    if len(sys.argv) > 2:
        np.savetxt(sys.argv[2],
                   np.concatenate([X, Y[:, None]], 1),
                   fmt=(["%.4f"] * X.shape[1]) + ["%d"])



import numpy as np


def meanvar_normalization(X, *Xtest):
    """Normalize features using moments.

    Linearly normalize each input feature to make it have zero mean
    and unit variance.  Test features, when given, are scaled using
    the statistics computed on X.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    Xtest : ndarray, shape (mtest, n) or None
         zero or more arrays of test features (one row per feature vector).

    Returns
    -------
    ndarray, shape (m, n)
        normalized features.
    ndarray, shape (mtest, n)
        normalized test features (one for each array in Xtest).

    """
    mu = X.mean(0, keepdims=True)
    sigma = X.std(0, keepdims=True)
    X = X - mu
    X /= np.maximum(sigma, 1e-15)  # 1e-15 avoids division by zero
    if not Xtest:
        return X
    Xtest = tuple((Xt - mu) / np.maximum(sigma, 1e-15) for Xt in Xtest)
    return (X,) + Xtest


def minmax_normalization(X, *Xtest):
    """Scale features in the [0, 1] range.

    Linearly normalize each input feature in the [0, 1] range.  Test
    features, when given, are scaled using the statistics computed on
    X.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    Xtest : ndarray, shape (mtest, n) or None
         zero or more arrays of test features (one row per feature vector).

    Returns
    -------
    ndarray, shape (m, n)
        normalized features.
    ndarray, shape (mtest, n)
        normalized test features (one for each array in Xtest).

    """
    xmin = X.min(0, keepdims=True)
    xmax = X.max(0, keepdims=True)
    X = X - xmin
    X /= np.maximum(xmax - xmin, 1e-15)  # 1e-15 avoids division by zero
    if not Xtest:
        return X
    Xtest = tuple((Xt - xmin) / np.maximum(xmax - xmin, 1e-15) for Xt in Xtest)
    return (X,) + Xtest


def maxabs_normalization(X, *Xtest):
    """Scale features in the [-1, 1] range.

    Linearly normalize each input feature in the [-1, 1] range by
    dividing them by the maximum absolute value.  Test features, when
    given, are scaled using the statistics computed on X.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    Xtest : ndarray, shape (mtest, n) or None
         zero or more arrays of test features (one row per feature vector).

    Returns
    -------
    ndarray, shape (m, n)
        normalized features.
    ndarray, shape (mtest, n)
        normalized test features (one for each array in Xtest).

    """
    # 1e-15 avoids division by zero
    amax = np.maximum(np.abs(X).max(0, keepdims=True), 1e-15)
    X = X / amax
    if not Xtest:
        return X
    Xtest = tuple(Xt / amax for Xt in Xtest)
    return (X,) + Xtest


def l2_normalization(X, *Xtest):
    """L2 normalization of feature vectors.

    Scale feature vectors to make it have unit Euclidean norm.  Test
    features, when given, are scaled as well.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    Xtest : ndarray, shape (mtest, n) or None
         zero or more arrays of test features (one row per feature vector).

    Returns
    -------
    ndarray, shape (m, n)
        normalized features.
    ndarray, shape (mtest, n)
        normalized test features (one for each array in Xtest).

    """
    q = np.sqrt((X ** 2).sum(1, keepdims=True))
    X = X / np.maximum(q, 1e-15)  # 1e-15 avoids division by zero
    if not Xtest:
        return X
    Xtest = tuple(l2_normalization(Xt) for Xt in Xtest)
    return (X,) + Xtest


def l1_normalization(X, *Xtest):
    """L1 normalization of feature vectors.

    Scale feature vectors to make it have unit L1 norm.  Test
    features, when given, are scaled as well.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    Xtest : ndarray, shape (mtest, n) or None
         zero or more arrays of test features (one row per feature vector).

    Returns
    -------
    ndarray, shape (m, n)
        normalized features.
    ndarray, shape (mtest, n)
        normalized test features (one for each array in Xtest).

    """
    q = np.abs(X).sum(1, keepdims=True)
    X = X / np.maximum(q, 1e-15)  # 1e-15 avoids division by zero
    if not Xtest:
        return X
    Xtest = tuple(l1_normalization(Xt) for Xt in Xtest)
    return (X,) + Xtest


def whitening(X, *Xtest):
    """Whitening transform.

    Linearly transform features to make it have zero mean, unit
    variance and null covariance.  Test features, when given, are
    trandformed using the statistics computed on X.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    Xtest : ndarray, shape (mtest, n) or None
         zero or more arrays of test features (one row per feature vector).

    Returns
    -------
    ndarray, shape (m, n)
        normalized features.
    ndarray, shape (mtest, n)
        normalized test features (one for each array in Xtest).

    """
    mu = X.mean(0, keepdims=True)
    sigma = np.cov(X.T)
    evals, evecs = np.linalg.eig(sigma)
    w = (np.maximum(evals, 1e-15) ** -0.5) * evecs  # 1e-15 avoids div. by zero
    X = (X - mu) @ w
    if not Xtest:
        return X
    Xtest = tuple((Xt - mu) @ w for Xt in Xtest)
    return (X,) + Xtest



import numpy as np


#!begin
def gradient_descent(loss_grad, theta, lr=0.001, maxsteps=100,
                     mindiff=1e-7):
    """Gradient descent optimization method.

    Find an approximation of the vector theta that minimizes loss_fun.

    Parameters
    ----------
    loss_grad : callable
        function computing the gradient.
    theta : ndarray, shape (n,)
        initial value.
    lr : float, optional
        learning rate.
    maxsteps : int, optional
        maximum number of iterations.
    mindiff : float, optional
        stop when the change is below this value.

    Returns
    -------
     ndarray, shape (n,)
        the final estimate found by the algorithm.

    """
    for _ in range(maxsteps):
        diff = lr * loss_grad(theta)
        theta -= diff
        if np.linalg.norm(diff) < mindiff:
            break
    return theta
#!end


#!begin2
def gradient_descent_with_momentum(loss_grad, theta, lr=0.001,
                                   momentum=0.9, maxsteps=100):
    """Gradient descent with momentum.

    Find an approximation of the vector theta that minimizes an
    objective function.

    Parameters
    ----------
    loss_grad : callable
        function computing the gradient of the objective.
    theta : ndarray, shape (n,)
        initial value.
    lr : float, optional
        learning rate.
    momentum : float, optional
        momentum term.
    maxsteps : int, optional
        maximum number of iterations.

    Returns
    -------
    ndarray, shape (n,)
        the final estimate found by the algorithm.

    """
    v = theta.copy()
    for _ in range(maxsteps):
        delta = momentum * v + loss_grad(theta)
        theta -= lr * delta
    return theta
#!end2


def check_gradient(f, grad_f, x, eps=1e-5):
    y = f(x)
    g = grad_f(x)
    y1 = f(x + eps * g)
    d = y1 - y - eps * (g ** 2).sum()
    print(d)


def _test():
    """Test the method."""
    target = np.array((1, 2, 3))

    def f(x):
        return ((x - target) ** 2).sum()

    def df(x):
        return 2 * (x - target)

    xx = gradient_descent(df, np.zeros(3), 0.1)
    print(xx)
    xx = gradient_descent_with_momentum(df, np.zeros(3), 0.1)
    print(xx)


if __name__ == "__main__":
    _test()



from .dataset import *
from .logistic_regression import *
from .svm import *
from .ksvm import *
from .multinomial_logistic import *
from .gda import *
from .naivebayes import *
from .nearest_neighbor import *
from .classtree import *
from .perceptron import *
from .mlp import *
from .cnn import *
from .normalization import *
from .pca import *
from .pvmlnet import *



import numpy as np
import pvml


def check_gradient(mlp, eps=1e-6, bsz=2):
    """Numerically check gradient computations."""
    insz = mlp.weights[0].shape[0]
    outsz = mlp.weights[-1].shape[1]
    X = np.random.randn(bsz, insz)
    Y = np.random.randint(0, outsz, (bsz,))
    A = mlp.forward(X)
    D = mlp.backward(Y, A)
    L = mlp.loss(Y, A[-1])
    for XX, W, DD in zip(A, mlp.weights, D):
        grad_W = (XX.T @ DD)
        GW = np.empty_like(W)
        for idx in np.ndindex(*W.shape):
            bak = W[idx]
            W[idx] += eps
            A1 = mlp.forward(X)
            L1 = mlp.loss(Y, A1[-1])
            W[idx] = bak
            GW[idx] = (L1 - L) / eps
        err = np.abs(GW - grad_W).max()
        print(err, "OK" if err < 1e-4 else "")
        assert err < 1e-4
    for b, DD in zip(mlp.biases, D):
        grad_b = DD.sum(0)
        Gb = np.empty_like(b)
        for idx in np.ndindex(*b.shape):
            bak = b[idx]
            b[idx] += eps
            A1 = mlp.forward(X)
            L1 = mlp.loss(Y, A1[-1])
            b[idx] = bak
            Gb[idx] = (L1 - L) / eps
        err = np.abs(Gb - grad_b).max()
        print(err, "OK" if err < 1e-4 else "")
        assert err < 1e-4


if __name__ == "__main__":
    mlp = pvml.MLP([3, 10])
    check_gradient(mlp)
    mlp = pvml.MLP([4, 7, 4])
    check_gradient(mlp)
    mlp = pvml.MLP([5, 6, 8, 7])
    check_gradient(mlp)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pvml",
    version="0.0.6",
    author="Claudio Cusano",
    author_email="claudio.cusano@unipv.it",
    description="A small and simple machine learning library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/claudio-unipv/pvml",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "numpy>=1.13.0",
        "matplotlib>=2.1.0",
    ],
    python_requires='>=3.6',
)


# python3 setup.py sdist bdist_wheel
# python3 -m twine upload dist/*



import numpy as np
from .multinomial_logistic import softmax


class MLP:
    """Multi-layer perceptron.

    A multi-layer perceptron for classification.

    The activation function for the output layer is the softmax
    operator, while hidden neurons use relu.  The loss function used
    during training is the cross entropy.

    To use a different architecture it is possible to define a new
    derived class which overrides some of the methods.

    """

    def __init__(self, neuron_counts):
        """Create and initialiaze the MLP.

        At least two layers must be specified (input and output).

        Parameters
        ----------
        neuron_counts : list
            number of neurons in the layers (first input, then hiddens,
            then output).
        """
        # Initialize weights with the Kaiming technique.
        self.weights = [np.random.randn(m, n) * np.sqrt(2.0 / m)
                        for m, n in zip(neuron_counts[:-1], neuron_counts[1:])]
        # Biases are zero-initialized.
        self.biases = [np.zeros(m) for m in neuron_counts[1:]]
        # Accumulators for the momentum terms
        self.update_w = [np.zeros_like(w) for w in self.weights]
        self.update_b = [np.zeros_like(b) for b in self.biases]

    def forward(self, X):
        """Compute the activations of the neurons.

        Parameters
        ----------
        X : ndarray, shape (m, n)
            input features (one row per feature vector).

        Returns
        -------
        list
            the list of activations of the neurons

        The returned list contains arrays (one per layer) with the
        activations of the layers.  The first element of the list is
        the input X, followed by the activations of hidden layers and
        trminated by the activations in the output layer.

        """
        activations = [X]
        for W, b in zip(self.weights[:-1], self.biases[:-1]):
            X = self.forward_hidden_layer(X, W, b)
            X = self.forward_hidden_activation(X)
            activations.append(X)
        X = self.forward_output_layer(X, self.weights[-1], self.biases[-1])
        X = self.forward_output_activation(X)
        activations.append(X)
        return activations

    def backward(self, Y, activations):
        """Compute the derivatives of the loss wrt the activations.

        Parameters
        ----------
        Y : ndarray, shape (m,)
            target output (integer class labels).
        activations : list
            activations computed by the forward method.

        Returns
        -------
        list
            the list of computed derivatives, one for each layer.
        """
        d = self.backward_output_activation(Y, activations[-1])
        derivatives = [d]
        if len(self.weights) > 1:
            d = self.backward_output_layer(self.weights[-1],
                                           self.biases[-1], d)
            d = self.backward_hidden_activation(activations[-2], d)
            derivatives.append(d)
        for W, b, X in zip(self.weights[-2:0:-1], self.biases[-2:0:-1],
                           activations[-3::-1]):
            d = self.backward_hidden_layer(W, b, d)
            d = self.backward_hidden_activation(X, d)
            derivatives.append(d)
        return derivatives[::-1]

    def backpropagation(self, X, Y, lr=1e-4, lambda_=1e-5, momentum=0.99):
        """Backpropagation algorithm.

        Perform both the forward and the backward steps and update the
        parameters.

        Parameters
        ----------
        X : ndarray, shape (m, n)
            input features (one row per feature vector).
        Y : ndarray, shape (m,)
            target output (integer class labels).
        lr : float
            learning rate.
        lambda_ : float
            regularization coefficients.
        momentum : float
            momentum coefficient.

        """
        activations = self.forward(X)
        derivatives = self.backward(Y, activations)
        for X, D, W, b, uw, ub in zip(activations, derivatives,
                                      self.weights, self.biases,
                                      self.update_w, self.update_b):
            grad_W = (X.T @ D) + lambda_ * W
            grad_b = D.sum(0)
            uw *= momentum
            uw -= lr * grad_W
            W += uw
            ub *= momentum
            ub -= lr * grad_b
            b += ub

    def inference(self, X):
        """Compute the predictions of the network.

        Parameters
        ----------
        X : ndarray, shape (m, n)
            input features (one row per feature vector).

        Returns
        -------
        ndarray, shape (m,)
            predicted labels, in the range 0, ..., k - 1
        ndarray, shape (m, k)
            posterior probability estimates.

        """
        probs = self.forward(X)[-1]
        labels = np.argmax(probs, 1)
        return labels, probs

    def train(self, X, Y, lr=1e-4, lambda_=1e-5, momentum=0.99,
              steps=10000, batch=None):
        """Train the network.

        Apply multiple steps of stochastic gradient descent.

        Parameters
        ----------
        X : ndarray, shape (m, n)
            input features (one row per feature vector).
        Y : ndarray, shape (m,)
            target output (integer class labels).
        lr : float
            learning rate.
        lambda_ : float
            regularization coefficients.
        momentum : float
            momentum coefficient.
        steps : int
            training iterations.
        batch : int or None
            size of the minibatch used in each step.  When None all
            the data is used in each step.

        """
        m = X.shape[0]
        if batch is None:
            batch = X.shape[0]
        i = m
        indices = np.arange(m)
        for step in range(steps):
            if i + batch > m:
                i = 0
                np.random.shuffle(indices)
            self.backpropagation(X[indices[i:i + batch], :],
                                 Y[indices[i:i + batch]],
                                 lr=lr,
                                 lambda_=lambda_,
                                 momentum=momentum)
            i += batch

    def save(self, filename):
        """Save the network to the file."""
        np.savez(filename, weights=self.weights, biases=self.biases)

    @classmethod
    def load(cls, filename):
        """Create a new network from the data saved in the file."""
        data = np.load(filename, allow_pickle=True)
        neurons = [w.shape[0] for w in data["weights"]]
        neurons.append(data["weights"][-1].shape[1])
        network = cls(neurons)
        network.weights = data["weights"]
        network.biases = data["biases"]
        return network

    # These last methods can be modified by derived classes to change
    # the architecture of the MLP.

    def forward_hidden_layer(self, X, W, b):
        """Forward pass of hidden layers."""
        return X @ W + b

    def forward_hidden_activation(self, X):
        """Activation function of hidden layers."""
        return relu(X)

    def forward_output_layer(self, X, W, b):
        """Forward pass of the output layer."""
        return X @ W + b

    def forward_output_activation(self, X):
        """Activation function of the output layer."""
        return softmax(X)

    def backward_hidden_layer(self, W, b, d):
        """Backward pass of hidden layers."""
        return d @ W.T

    def backward_hidden_activation(self, Y, d):
        """Derivative of the activation function of hidden layers."""
        return d * (Y > 0).astype(int)

    def backward_output_layer(self, W, b, d):
        """Backward pass of the ouput layer."""
        return d @ W.T

    def backward_output_activation(self, Y, P):
        """Derivative of the activation function of output layer."""
        d = P.copy()
        # Implicitly subtract the one-hot vectors
        d[np.arange(Y.shape[0]), Y] -= 1
        return d / Y.shape[0]

    def loss(self, Y, P):
        """Compute the average cross-entropy."""
        return -np.log(P[np.arange(Y.shape[0]), Y]).mean()


def relu(x):
    """ReLU activation function."""
    return np.maximum(x, 0)



import numpy as np


def logreg_inference(X, w, b):
    """Predict class probabilities.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    w : ndarray, shape (n,)
         weight vector.
    b : float
         scalar bias.

    Returns
    -------
    ndarray, shape (m,)
        probability estimates (one per feature vector).
    """
    logits = X @ w + b
    return 1 / (1 + np.exp(-logits))


def binary_cross_entropy(Y, P):
    """Average cross entropy.

    Parameters
    ----------
    Y : ndarray, shape (m,)
        binary target labels (0 or 1).
    P : ndarray, shape (m,)
        probability estimates.

    Returns
    -------
    float
        average cross entropy.
    """
    eps = 1e-3
    P = np.clip(P, eps, 1 - eps)  # This prevents overflows
    return -(Y * np.log(P) + (1 - Y) * np.log(1 - P)).mean()


def logreg_train(X, Y, lr=1e-3, steps=1000, init_w=None, init_b=0):
    """Train a binary classifier based on logistic regression.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features.
    Y : ndarray, shape (m,)
        binary training labels.
    lr : float
        learning rate
    steps : int
        number of training steps
    init_w : ndarray, shape (n,)
        initial weights (None for zero initialization)
    init_b : float
        initial bias

    Returns
    -------
    w : ndarray, shape (n,)
        learned weight vector.
    b : float
        learned bias.
    """
    m, n = X.shape
    w = (init_w if init_w is not None else np.zeros(n))
    b = init_b
    for step in range(steps):
        P = logreg_inference(X, w, b)
        grad_w = ((P - Y) @ X) / m
        grad_b = (P - Y).mean()
        w -= lr * grad_w
        b -= lr * grad_b
    return w, b


def logreg_l2_train(X, Y, lambda_, lr=1e-3, steps=1000, init_w=None,
                    init_b=0):
    """Train a binary classifier based on L2-regularized logistic regression.

    Parameters
    ----------

    X : ndarray, shape (m, n)
        training features.
    Y : ndarray, shape (m,)
        binary training labels.
    lambda_ : float
        regularization coefficient.
    lr : float
        learning rate.
    steps : int
        number of training steps.
    init_w : ndarray, shape (n,)
        initial weights (None for zero initialization)
    init_b : float
        initial bias

    Returns
    -------

    w : ndarray, shape (n,)
        learned weight vector.
    b : float
        learned bias.
    """
    m, n = X.shape
    w = (init_w if init_w is not None else np.zeros(n))
    b = init_b
    for step in range(steps):
        P = logreg_inference(X, w, b)
        grad_w = ((P - Y) @ X) / m + 2 * lambda_ * w
        grad_b = (P - Y).mean()
        w -= lr * grad_w
        b -= lr * grad_b
    return w, b


def logreg_l1_train(X, Y, lambda_, lr=1e-3, steps=1000, init_w=None, init_b=0):
    """Train a binary classifier based on L1-regularized logistic regression.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features.
    Y : ndarray, shape (m,)
        binary training labels.
    lambda_ : float
        regularization coefficient.
    lr : float
        learning rate.
    steps : int
        number of training steps.
    loss : ndarray, shape (steps,)
        loss value after each training step.

    Returns
    -------
    w : ndarray, shape (n,)
        learned weight vector.
    b : float
        learned bias.
    """
    m, n = X.shape
    w = (init_w if init_w is not None else np.zeros(n))
    b = init_b
    for step in range(steps):
        P = logreg_inference(X, w, b)
        grad_w = ((P - Y) @ X) / m + lambda_ * np.sign(w)
        grad_b = (P - Y).mean()
        w -= lr * grad_w
        b -= lr * grad_b
    return w, b



import numpy as np
import svm


#!begin1
def one_vs_rest_svm_train(X, Y, lambda_, lr=1e-3, steps=1000):
    """Train a multi-class linear SVM using the one vs. rest strategy.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features.
    Y : ndarray, shape (m,)
        training labels in the range 0, ..., (k - 1).
    lambda_ : float
        regularization coefficient.
    lr : float
        learning rate
    steps : int
        number of training steps

    Returns
    -------
    w : ndarray, shape (n, k)
        learned weights (one vector per class).
    b : ndarray, shape (k,)
        vector of biases.
    loss : ndarray, shape (steps,)
        loss value after each training iteration.
    """
    k = Y.max() + 1
    m, n = X.shape
    W = np.zeros((n, k))
    b = np.zeros(k)
    loss = np.empty((k, steps))
    for c in range(k):
        Ybin = (Y == c)
        Wbin, bbin, ls = svm.svm_train(X, Ybin, lambda_, lr=lr, steps=steps)
        W[:, c] = Wbin
        b[c] = bbin
        loss[c, :] = ls
    return W, b, loss.mean(0)


def one_vs_rest_svm_inference(X, W, b):
    """Multiclass linear SVM prediction of the class labels.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    w : ndarray, shape (n, k)
         weights (one vector per class).
    b : ndarray, shape (k,)
         vector of biases.

    Returns
    -------
    ndarray, shape (m,)
        predicted labels (one per feature vector) in the range 0...(k-1).
    ndarray, shape (m, k)
        classification scores.
    """
    logits = X @ W + b.T
    labels = np.argmax(logits, 1)
    return labels, logits
#!end1


#!begin2
def one_vs_one_svm_train(X, Y, lambda_, lr=1e-3, steps=1000):
    """Train a multi-class linear SVM using the one vs. one strategy.

    Parameters
    ----------
    X : ndarray, shape (m, n)
        training features.
    Y : ndarray, shape (m,)
        training labels in the range 0, ..., (k - 1).
    lambda_ : float
        regularization coefficient.
    lr : float
        learning rate
    steps : int
        number of training steps

    Returns
    -------
    w : ndarray, shape (n, k * (k - 1) // 2)
        learned weights (one vector per each pair of classes).
    b : ndarray, shape (k * (k - 1) // 2,)
        vector of biases.
    loss : ndarray, shape (steps,)
        loss value after each training iteration.
    """
    k = Y.max() + 1
    m, n = X.shape
    W = []
    b = []
    loss = []
    # For each pair of classes...
    for pos in range(k):
        for neg in range(pos + 1, k):
            # Build a training subset
            subset = (np.logical_or(Y == pos, Y == neg)).nonzero()[0]
            Xbin = X[subset, :]
            Ybin = (Y[subset] == pos)
            # Train the classifier
            Wbin, bbin, ls = svm.svm_train(Xbin, Ybin, lambda_, lr=lr,
                                           steps=steps)
            W.append(Wbin)
            b.append(bbin)
            loss.append(ls)
    # Convert lists into arrays
    W = np.vstack(W).T
    b = np.stack(b)
    loss = np.vstack(loss).mean(0)
    return W, b, loss


def one_vs_one_svm_inference(X, W, b):
    """Multiclass linear SVM prediction of the class labels.

    Parameters
    ----------
    X : ndarray, shape (m, n)
         input features (one row per feature vector).
    w : ndarray, shape (n, s)
         weights (one vector per pair of classes).
    b : ndarray, shape (s,)
         vector of biases.

    Returns
    -------
    ndarray, shape (m,)
        predicted labels (one per feature vector) in the range 0...(k-1).
    ndarray, shape (m, k)
        classification scores.
    """
    # 1) recover the number of classes from s = 1 + 2 + ... + k
    m = X.shape[0]
    s = b.size
    k = int(1 + np.sqrt(1 + 8 * s)) // 2
    votes = np.zeros((m, k))
    logits = X @ W + b.T
    bin_labels = (logits > 0)
    # For each pair of classes...
    j = 0
    for pos in range(k):
        for neg in range(pos + 1, k):
            votes[:, pos] += bin_labels[:, j]
            votes[:, neg] += (1 - bin_labels[:, j])
            j += 1
    labels = np.argmax(votes, 1)
    return labels, votes
#!end2


if __name__ == "__main__":
    import demo

    class Demo(demo.Demo):
        def train(self, X, Y):
            if self.args.strategy == "ova":
                w, b, loss = one_vs_rest_svm_train(X, Y,
                                                   self.args.lambda_,
                                                   lr=self.args.lr,
                                                   steps=self.args.steps)
            else:
                w, b, loss = one_vs_one_svm_train(X, Y,
                                                  self.args.lambda_,
                                                  lr=self.args.lr,
                                                  steps=self.args.steps)
            self.w = w
            self.b = b
            return loss

        def inference(self, X):
            if self.args.strategy == "ova":
                return one_vs_rest_svm_inference(X, self.w, self.b)[0]
            else:
                return one_vs_one_svm_inference(X, self.w, self.b)[0]

        def report(self, X, *args):
            super().report(X, *args)
            if X.shape[1] == 2 and self.args.draw_planes:
                for i in range(self.w.shape[1]):
                    demo.draw_linear_boundary(self.w[:, i], self.b[i],
                                              linestyle="--")

    demoapp = Demo()
    demoapp.parser.add_argument("--strategy", choices=["ova", "ovo"],
                                default="ova",
                                help="multiclass strategy")
    demoapp.parser.add_argument("--draw-planes", action="store_true",
                                help="draw hyperplanes")
    demoapp.run()
